from typing import Callable

import numpy as np
import scipy.optimize


def minimize(fun: Callable, x0, *args, **kwargs):
    x0 = np.asarray(x0)
    x0_shape = x0.shape

    def fwrap(x):
        x = x.reshape(x0_shape)
        val = fun(x)
        val_shape = np.asarray(val).shape
        if val_shape != ():
            raise ValueError(
                "Objective function must return a true scalar. "
                f"Got array of shape {val_shape}."
            )
        return val

    out = scipy.optimize.minimize(fwrap, x0, *args, **kwargs)
    out.x = out.x.reshape(x0.shape)
    return out


def leastsq(fun: Callable, x0, *args, **kwargs):
    x0 = np.asarray(x0)
    x0_shape = x0.shape

    def fwrap(x):
        x = x.reshape(x0_shape)
        return fun(x)

    out = scipy.optimize.leastsq(fwrap, x0, *args, **kwargs)
    return (out[0].reshape(x0.shape), *out[1:])



from typing import Callable, Optional

import numpy as np


def _dot_last(a, b):
    a_shape = a.shape
    a = a.reshape(*a.shape[:-1], -1)
    b = b.reshape(*b.shape[:-1], -1)
    out = np.dot(a.T, b)
    out = out.reshape(a_shape[:-1])
    return out


def bisect(f: Callable, a, b, tol: float, max_num_steps: Optional[int] = None):
    a = np.asarray(a)
    b = np.asarray(b)

    fa = np.asarray(f(a))
    fb = np.asarray(f(b))

    assert np.all(np.logical_xor(fa > 0, fb > 0))
    is_fa_positive = fa > 0
    # sort points such that f(a) is negative, f(b) positive
    tmp = fa[is_fa_positive]
    fa[is_fa_positive] = fb[is_fa_positive]
    fb[is_fa_positive] = tmp
    #
    tmp = a[..., is_fa_positive]
    a[..., is_fa_positive] = b[..., is_fa_positive]
    b[..., is_fa_positive] = tmp

    k = 0
    while True:
        diff = a - b
        dist2 = _dot_last(diff, diff)

        if np.all(dist2 < tol ** 2):
            break

        if max_num_steps is not None and k >= max_num_steps:
            break

        c = (a + b) / 2
        fc = np.asarray(f(c))

        is_fc_positive = fc > 0
        a[..., ~is_fc_positive] = c[..., ~is_fc_positive]
        b[..., is_fc_positive] = c[..., is_fc_positive]

        k += 1

    return a, b


def regula_falsi(f: Callable, a, b, tol: float, max_num_steps: Optional[int] = None):
    a = np.asarray(a)
    b = np.asarray(b)
    fa = np.asarray(f(a))
    fb = np.asarray(f(b))

    assert np.all(np.logical_xor(fa > 0, fb > 0))
    # sort points such that f(a) is negative, f(b) positive
    is_fa_positive = fa > 0
    fa[is_fa_positive], fb[is_fa_positive] = fb[is_fa_positive], fa[is_fa_positive]
    a[..., is_fa_positive], b[..., is_fa_positive] = (
        b[..., is_fa_positive],
        a[..., is_fa_positive],
    )

    k = 0
    while True:
        if max_num_steps is not None and k >= max_num_steps:
            break

        c = (a * fb - b * fa) / (fb - fa)
        fc = f(c)

        is_fc_positive = fc > 0
        a[~is_fc_positive] = c[~is_fc_positive]
        b[is_fc_positive] = c[is_fc_positive]

        diff = a - b
        dist2 = _dot_last(diff, diff)

        if np.all(dist2 < tol ** 2):
            break

        k += 1

    return a, b



from collections import namedtuple
from typing import Callable, Optional

import numpy as np
import scipy
import scipy.sparse.linalg

Info = namedtuple("KrylovInfo", ["success", "xk", "numsteps", "resnorms", "errnorms"])


def _norm(r, M=None):
    Mr = r if M is None else M @ r

    if len(r.shape) == 1:
        rMr = np.dot(r.conj(), Mr)
    else:
        rMr = np.einsum("i...,i...->...", r.conj(), Mr)

    if np.any(rMr.imag) > 1.0e-13:
        raise RuntimeError(
            "Got nonnegative imaginary part. " "Is the preconditioner not self-adjoint?"
        )
    return np.sqrt(rMr.real)


def assert_shapes(A, b, x0):
    if len(A.shape) != 2 or A.shape[0] != A.shape[1]:
        raise ValueError(f"A must be a square matrix, not A.shape = {A.shape}.")
    if x0.shape != b.shape:
        raise ValueError(
            f"x0 and b need to have the same shape, not x0.shape = {x0.shape}, "
            f"b.shape = {b.shape}"
        )
    if b.shape[0] != b.size:
        raise ValueError("Can only deal with one right-hand side at a time.")
    assert A.shape[1] == b.shape[0]


def _wrapper(
    method,
    A,
    b,
    x0=None,
    tol: float = 1e-05,
    maxiter: Optional[int] = None,
    M=None,
    callback: Optional[Callable] = None,
    atol: Optional[float] = 0.0,
    exact_solution=None,
):
    if x0 is None:
        x0 = np.zeros_like(b)

    assert_shapes(A, b, x0)

    x0_shape = x0.shape

    # initial residual
    r = b - A @ x0
    resnorms = [_norm(r, M)]

    if exact_solution is None:
        errnorms = None
    else:
        err = exact_solution - x0
        errnorms = [_norm(err)]

    num_steps = 0

    def cb(xk):
        nonlocal num_steps
        num_steps += 1

        xk = xk.reshape(x0_shape)

        if callback is not None:
            callback(xk)

        r = b - A @ xk
        resnorms.append(_norm(r, M))

        if exact_solution is not None:
            errnorms.append(_norm(exact_solution - x0))

    x, info = method(A, b, x0=x0, tol=tol, maxiter=maxiter, M=M, atol=atol, callback=cb)
    success = info == 0
    x = x.reshape(x0_shape)
    return x if success else None, Info(success, x, num_steps, resnorms, errnorms)


def cg(*args, **kwargs):
    return _wrapper(scipy.sparse.linalg.cg, *args, **kwargs)


def bicg(*args, **kwargs):
    return _wrapper(scipy.sparse.linalg.bicg, *args, **kwargs)


def bicgstab(*args, **kwargs):
    return _wrapper(scipy.sparse.linalg.bicgstab, *args, **kwargs)


def cgs(*args, **kwargs):
    return _wrapper(scipy.sparse.linalg.cgs, *args, **kwargs)


def gmres(
    A,
    b,
    x0=None,
    tol: float = 1e-05,
    restart: Optional[int] = None,
    maxiter: Optional[int] = None,
    M=None,
    callback: Optional[Callable] = None,
    atol: Optional[float] = 0.0,
    exact_solution=None,
):
    if x0 is None:
        x0 = np.zeros(A.shape[1])

    assert_shapes(A, b, x0)

    x0_shape = x0.shape

    # scipy.gmres() apparently calls the callback before the start of the iteration such
    # that the initial residual is automatically contained
    resnorms = []
    num_steps = -1

    if exact_solution is None:
        errnorms = None
    else:
        errnorms = []

    def cb(xk):
        nonlocal num_steps
        num_steps += 1

        xk = xk.reshape(x0_shape)

        if callback is not None:
            callback(xk)

        r = b - A @ xk
        resnorms.append(_norm(r, M))

        if exact_solution is not None:
            errnorms.append(_norm(exact_solution - x0))

    x, info = scipy.sparse.linalg.gmres(
        A,
        b,
        x0=x0,
        tol=tol,
        restart=restart,
        maxiter=maxiter,
        M=M,
        atol=atol,
        callback=cb,
        callback_type="x",
    )
    success = info == 0
    x = x.reshape(x0_shape)
    return x if success else None, Info(success, x, num_steps, resnorms, errnorms)


# Need a special minres wrapper since it doesn't have atol.
# <https://github.com/scipy/scipy/issues/13935>
def minres(
    A,
    b,
    x0=None,
    shift: float = 0.0,
    tol: float = 1e-05,
    maxiter: Optional[int] = None,
    M=None,
    callback: Optional[Callable] = None,
    exact_solution=None,
):
    if x0 is None:
        x0 = np.zeros(A.shape[1])

    assert_shapes(A, b, x0)

    x0_shape = x0.shape

    # initial residual
    resnorms = []
    r = b - A @ x0
    resnorms.append(_norm(r, M))

    if exact_solution is None:
        errnorms = None
    else:
        err = exact_solution - x0
        errnorms = [_norm(err)]

    num_steps = 0

    def cb(xk):
        nonlocal num_steps
        num_steps += 1

        xk = xk.reshape(x0_shape)

        if callback is not None:
            callback(xk)

        r = b - A @ xk
        resnorms.append(_norm(r, M))

        if exact_solution is not None:
            errnorms.append(_norm(exact_solution - x0))

    x, info = scipy.sparse.linalg.minres(
        A, b, x0=x0, shift=shift, tol=tol, maxiter=maxiter, M=M, callback=cb
    )
    success = info == 0
    x = x.reshape(x0_shape)
    return x if success else None, Info(success, x, num_steps, resnorms, errnorms)


# treat qmr separately because it can deal with two preconditioners, M1 and M2 (left and
# right)
def qmr(
    A,
    b,
    x0=None,
    tol: float = 1e-05,
    maxiter: Optional[int] = None,
    M1=None,
    M2=None,
    callback: Optional[Callable] = None,
    atol: Optional[float] = 0.0,
    exact_solution=None,
):
    if x0 is None:
        x0 = np.zeros(A.shape[1])

    x0_shape = x0.shape

    assert_shapes(A, b, x0)

    # initial residual
    resnorms = []
    r = b - A @ x0
    resnorms.append(_norm(r, M1))

    if exact_solution is None:
        errnorms = None
    else:
        err = exact_solution - x0
        errnorms = [_norm(err)]

    num_steps = 0

    def cb(xk):
        nonlocal num_steps
        num_steps += 1

        xk = xk.reshape(x0_shape)

        if callback is not None:
            callback(xk)

        r = b - A @ xk
        resnorms.append(_norm(r, M1))

        if exact_solution is not None:
            errnorms.append(_norm(exact_solution - x0))

    x, info = scipy.sparse.linalg.qmr(
        A, b, x0=x0, tol=tol, maxiter=maxiter, M1=M1, M2=M2, atol=atol, callback=cb
    )
    success = info == 0
    x = x.reshape(x0_shape)
    return x if success else None, Info(success, x, num_steps, resnorms, errnorms)



import numpy as np
import pytest
import scipy.sparse.linalg

import scipyx


def _spd(shape_rhs):
    n = shape_rhs[0]
    data = -np.ones((3, n))
    data[1] = 2.0
    A = scipy.sparse.spdiags(data, [-1, 0, 1], n, n)
    A = A.tocsr()
    b = np.ones(shape_rhs)
    return A, b, None


def _spd_prec():
    n = 5
    data = -np.ones((3, n))
    data[1] = 2.0
    A = scipy.sparse.spdiags(data, [-1, 0, 1], n, n)
    A = A.tocsr()
    b = np.ones(n)
    #
    M = scipy.sparse.linalg.LinearOperator(
        (n, n), matvec=lambda x: 0.5 * x, rmatvec=lambda x: 0.5 * x
    )
    return A, b, M


def _hpd():
    a = np.array(np.linspace(1.0, 2.0, 5), dtype=complex)
    a[0] = 5.0
    a[-1] = 1.0e-1
    A = np.diag(a)
    A[-1, 0] = 1.0e-1j
    A[0, -1] = -1.0e-1j

    b = np.ones(5, dtype=complex)
    return A, b, None


@pytest.mark.parametrize(
    "method, system",
    [
        (scipyx.cg, _spd((5,))),
        (scipyx.cg, _spd((5, 1))),
        (scipyx.cg, _spd_prec()),
        (scipyx.cg, _hpd()),
        #
        (scipyx.gmres, _spd((5,))),
        (scipyx.gmres, _spd((5, 1))),
        (scipyx.gmres, _spd_prec()),
        (scipyx.gmres, _hpd()),
        #
        (scipyx.minres, _spd((5,))),
        (scipyx.minres, _spd((5, 1))),
        (scipyx.minres, _spd_prec()),
        # (scipyx.minres, _hpd()),  ERR minres can't deal with hermitian?
        #
        (scipyx.bicg, _spd((5,))),
        (scipyx.bicg, _spd((5, 1))),
        (scipyx.bicg, _spd_prec()),
        (scipyx.bicg, _hpd()),
        #
        (scipyx.bicgstab, _spd((5,))),
        (scipyx.bicgstab, _spd((5, 1))),
        (scipyx.bicgstab, _spd_prec()),
        (scipyx.bicgstab, _hpd()),
        #
        (scipyx.cgs, _spd((5,))),
        (scipyx.cgs, _spd((5, 1))),
        (scipyx.cgs, _spd_prec()),
        (scipyx.cgs, _hpd()),
        #
        (scipyx.qmr, _spd((5,))),
        (scipyx.qmr, _spd((5, 1))),
        # (scipyx.qmr, _spd_prec()),
        (scipyx.qmr, _hpd()),
    ],
)
def test_run(method, system, tol=1.0e-13):
    A, b, M = system

    if isinstance(A, np.ndarray):
        exact_solution = np.linalg.solve(A, b)
    else:
        exact_solution = np.linalg.solve(A.toarray(), b)

    def cb(x):
        assert x.shape == b.shape

    x0 = np.zeros_like(b)
    if M is None:
        sol, info = method(A, b, x0, exact_solution=exact_solution, callback=cb)
    else:
        sol, info = method(A, b, x0, exact_solution=exact_solution, M=M, callback=cb)
    assert sol is not None
    assert info.success
    assert len(info.resnorms) == info.numsteps + 1
    assert len(info.errnorms) == info.numsteps + 1

    assert np.all(np.abs(sol - info.xk) < tol * (1.0 + np.abs(sol)))
    assert sol.shape == x0.shape

    print(info.resnorms)
    assert np.asarray(info.resnorms).dtype == float
    assert np.asarray(info.errnorms).dtype == float

    # make sure the initial resnorm and errnorm are correct
    print("a")
    print(method)
    print(info.resnorms[0])
    if M is None:
        assert abs(np.linalg.norm(A @ x0 - b) - info.resnorms[0]) < tol
    else:
        r = A @ x0 - b
        resnorm0 = np.sqrt(np.dot(r, M @ r))
        assert abs(resnorm0 - info.resnorms[0]) < 1.0e-13
    assert abs(np.linalg.norm(x0 - exact_solution) - info.errnorms[0]) < tol



import numpy as np
import pytest

import scipyx as spx


def test_0d():
    def f(x):
        return (x ** 2 - 2) ** 2

    x0 = 1.5
    out = spx.minimize(f, x0)

    assert out.x.shape == np.asarray(x0).shape
    assert np.asarray(out.fun).shape == ()


def test_2d():
    def f(x):
        return (np.sum(x ** 2) - 2) ** 2

    x0 = np.ones((4, 3), dtype=float)
    out = spx.minimize(f, x0, method="Powell")

    assert out.x.shape == np.asarray(x0).shape
    assert np.asarray(out.fun).shape == ()


def test_error():
    def f(x):
        return x - 2

    x0 = [1.5]
    with pytest.raises(ValueError):
        spx.minimize(f, x0)


def test_0d_leastsq():
    def f(x):
        assert x.shape == ()
        return x

    x0 = 0.0
    x, _ = spx.leastsq(f, x0)
    assert x.shape == ()


def test_1d_leastsq():
    def f(x):
        assert x.shape == (2,)
        print(x.shape)
        return x

    x0 = [1.0, 0.0]
    x, _ = spx.leastsq(f, x0)
    assert x.shape == (2,)



import numpy as np

import scipyx


def test_bisect():
    def f(x):
        return x ** 2 - 2

    tol = 1.0e-12
    a, b = scipyx.bisect(f, 0.0, 5.0, tol)
    assert b - a < tol
    assert abs(np.sqrt(2) - a) < 2 * tol


def test_regula_falsi():
    def f(x):
        return x ** 2 - 2

    tol = 1.0e-12
    a, b = scipyx.regula_falsi(f, 0.0, 5.0, tol)
    assert b - a < tol
    assert abs(np.sqrt(2) - a) < 2 * tol
